/*
   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by BoLiu <Bo@cirrus.com>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
#ifndef __TCC893x_VIOC__
#define __TCC893x_VIOC__
#include <dfb_types.h>
#include <core/coretypes.h>
#include <core/layers.h>

/*
 * The following are the IOCTLs that can be sent to the EP93xx frame buffer
 * device.
 */
#define TCC_DFB_PATH_CONNECT		0x100
#define TCC_DFB_VIOC_IOCTL			0x101

typedef enum{
	TCC_LCDC_IMG_FMT_1BPP,
	TCC_LCDC_IMG_FMT_2BPP,
	TCC_LCDC_IMG_FMT_4BPP,
	TCC_LCDC_IMG_FMT_8BPP,
	TCC_LCDC_IMG_FMT_RGB332 = 8,
	TCC_LCDC_IMG_FMT_RGB444 = 9,
	TCC_LCDC_IMG_FMT_RGB565 = 10,
	TCC_LCDC_IMG_FMT_RGB555 = 11,
	TCC_LCDC_IMG_FMT_RGB888 = 12,
	TCC_LCDC_IMG_FMT_RGB666 = 13,
	TCC_LCDC_IMG_FMT_RGB888_3	= 14,		/* RGB888 - 3bytes aligned - B1[31:24],R0[23:16],G0[15:8],B0[7:0] newly supported : 3 bytes format*/
	TCC_LCDC_IMG_FMT_ARGB6666_3 = 15,		/* ARGB6666 - 3bytes aligned - A[23:18],R[17:12],G[11:6],B[5:0]newly supported : 3 bytes format */
	TCC_LCDC_IMG_FMT_COMP = 16,				// 4bytes
	TCC_LCDC_IMG_FMT_DECOMP	= (TCC_LCDC_IMG_FMT_COMP),
	TCC_LCDC_IMG_FMT_444SEP = 21,
	TCC_LCDC_IMG_FMT_UYVY = 22,		/* 2bytes	: YCbCr 4:2:2 Sequential format LSB [Y/U/Y/V] MSB : newly supported : 2 bytes format*/
	TCC_LCDC_IMG_FMT_VYUY = 23,		/* 2bytes	: YCbCr 4:2:2 Sequential format LSB [Y/V/Y/U] MSB : newly supported : 2 bytes format*/

	TCC_LCDC_IMG_FMT_YUV420SP = 24,	
	TCC_LCDC_IMG_FMT_YUV422SP = 25, 
	TCC_LCDC_IMG_FMT_YUYV = 26, 
	TCC_LCDC_IMG_FMT_YVYU = 27,
	
	TCC_LCDC_IMG_FMT_YUV420ITL0 = 28, 
	TCC_LCDC_IMG_FMT_YUV420ITL1 = 29, 
	TCC_LCDC_IMG_FMT_YUV422ITL0 = 30, 
	TCC_LCDC_IMG_FMT_YUV422ITL1 = 31, 
	TCC_LCDC_IMG_FMT_MAX
}TCC_LCDC_IMG_FMT_TYPE;

typedef enum {
	SCALER_POLLING,
	SCALER_INTERRUPT,
	SCALER_NOWAIT
} Scaler_ResponseType;

typedef struct {
	FBDev		*dfb_fbdev;
} VIOCDriverData;

typedef struct {
	unsigned int 			y_addr;	
	unsigned int 			u_addr;	
	unsigned int 			v_addr;	
	unsigned int 			format; 	

	unsigned int 			posx; 	
	unsigned int 			posy;
	
	unsigned int 			width; 	
	unsigned int 			height;

	unsigned int 			bpp;
	unsigned int 			pitch;	
}VIOC_XDMA_PARAM;

typedef struct {
	Scaler_ResponseType responsetype; 	//scaler response type
	unsigned int		src_Yaddr;	// destination image address
	unsigned int 		src_Uaddr;	// destination image address
	unsigned int 		src_Vaddr;	// destination image address
	unsigned int 		src_fmt;			// source image format
	unsigned int  		src_ImgWidth;	// source image width
	unsigned int  		src_ImgHeight;	// source image height
	unsigned int 		src_winLeft;
	unsigned int		src_winTop;
	unsigned int		src_winRight;
	unsigned int		src_winBottom;
	unsigned int		src_offset;
	unsigned int		dest_Yaddr;	// destination image address
	unsigned int 		dest_Uaddr;	// destination image address
	unsigned int 		dest_Vaddr;	// destination image address
	unsigned int 		dest_fmt;		// destination image format
	unsigned int  		dest_ImgWidth;	// destination image width
	unsigned int  		dest_ImgHeight;	// destination image height
	unsigned int		dest_winLeft;
	unsigned int		dest_winTop;
	unsigned int		dest_winRight;
	unsigned int		dest_winBottom;
	unsigned int		dest_offset;
	unsigned	int		r;
	unsigned	int		g;
	unsigned	int		b;
	unsigned	int		a;
	unsigned int		blit_status;
	unsigned int 		viqe_onthefly; 	// 0 : m to m , 0x1 : onthefly read  , 0x2 : onthefly write , 0x3 : read & write onthefly
	unsigned int		interlaced;
	unsigned char 	wdma_r2y; 		// 0 : No Operation, 1: WDMA R2Y Operation
	bool				bColorkey;
	bool				bAlpha;
}DFB_TYPE;

typedef struct {
	unsigned long fb_addr;
	u32 fill_color;
	u32 pixelformat;
	u8  pixeldepth;
	bool fb_store;
	unsigned long srcaddr,destaddr,srcpitch,destpitch;
	DFBRegion clip;
     /* state validation */
        int smf_source;
        int smf_destination;
        int smf_color;
	int smf_clip;
	
	unsigned int responsetype;
	unsigned int vioc_phy;
	VIOC_XDMA_PARAM src;
	VIOC_XDMA_PARAM dst;
	unsigned int dfb_fd;
	DFB_TYPE	 tcc_dfb_dev;
} VIOCDeviceData;



#endif /*__TCC893x_VIOC__*/
